
#include <errno.h>
#include <getopt.h>
#include <paths.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>

#define VERSION "1.0"

#define PID_FILE _PATH_VARRUN "vpctimed.pid"

#define FALSE 0
#define TRUE  1

/* Default values for options. */
#define DEFAULT_NEG_THRESHOLD	(-5)
#define DEFAULT_POS_THRESHOLD	(5)
#define DEFAULT_INTERVAL	(10)

/* Program name */
char *progname;

/* Options */
int    opt_debug         = FALSE;
int    opt_interval      = DEFAULT_INTERVAL;
double opt_neg_threshold = DEFAULT_NEG_THRESHOLD;
double opt_pos_threshold = DEFAULT_POS_THRESHOLD;

/*
 * get_host_time
 *
 * Get the current time of day from the host operating system.
 */
inline void get_host_time(int *seconds, int *minutes, int *hour)
{
    // Virtual PC-specific opcode to get host time
    asm(".byte 0x0f, 0x3f, 0x03, 0x03" :
	"=b" (*seconds), "=c" (*minutes), "=d" (*hour));
}

/*
 * get_host_date
 *
 * Get the current date from the host operating system.
 */
inline void get_host_date(int *day, int *month, int *year)
{
    // Virtual PC-specific opcode to get host date
    asm(".byte 0x0f, 0x3f, 0x03, 0x04" :
	"=b" (*day), "=c" (*month), "=d" (*year));
}

/*
 * handle_sigill
 *
 * Handle an illegal instruction signal.  We use some instructions that are
 * specific to Virtual PC; on a real processor these will generate a SIGILL.
 */
void handle_sigill(int signum)
{
    fprintf(stderr, "The processor does not seem to support Virtual PC-specific opcodes.\n");
    fprintf(stderr, "This program can only be run under Connectix Virtual PC.\n");
    exit(1);
}

/*
 * cleanup_and_exit
 *
 * Clean up and exit with the given exit code.
 */
void cleanup_and_exit(int exit_code)
{
    if (!opt_debug) {
	unlink(PID_FILE);
	closelog();
    }
    exit(exit_code);
}

/*
 * handle_signals
 *
 * Handle SIGINT, SIGQUIT, or SIGTERM by exiting gracefully.
 */
void handle_signals(int signum)
{
    if (!opt_debug)
	syslog(LOG_INFO, "Exiting");

    cleanup_and_exit(0);
}

/*
 * hosttime
 *
 * Return the time since the Epoch (00:00:00 UTC, January 1, 1970)
 * according to the host operating system's clock.
 */
time_t hosttime()
{
    struct tm host_tm;
    host_tm.tm_isdst = -1; // Daylight savings time = unknown

    get_host_time(&host_tm.tm_sec, &host_tm.tm_min, &host_tm.tm_hour);
    get_host_date(&host_tm.tm_mday, &host_tm.tm_mon, &host_tm.tm_year);

    host_tm.tm_mon  -= 1;    // tm_mon is months since January
    host_tm.tm_year -= 1900; // tm_year is years after 1900

    return mktime(&host_tm);
}

/*
 * synctime
 *
 * Check the difference between the guest time and host time, and reset
 * the guest time if it is wrong by more than the threshold given by
 * opt_neg_threshold and opt_pos_threshold.
 */
void synctime()
{
    struct timeval old_tv, new_tv;
    double diff;

    /* Get the current time according to the guest Linux/Unix OS */
    gettimeofday(&old_tv, NULL);

    /* Get the current time according to the host. */
    new_tv.tv_sec  = hosttime();
    new_tv.tv_usec = 0;

    /*
     * If the time is off by more than the threshold, set it to the host time.
     */
    diff = difftime(new_tv.tv_sec, old_tv.tv_sec);    

    if (opt_debug) {
	putchar('\n');
	printf(" Host time: %s", ctime(&new_tv.tv_sec));
	printf("Guest time: %s", ctime(&old_tv.tv_sec));
	printf("Difference: %.0f seconds\n", diff);
    }

    if (diff < opt_neg_threshold || diff > opt_pos_threshold) {
	if (opt_debug)
	    printf("Synchronizing guest time.\n");

	if (settimeofday(&new_tv, NULL) != 0) {
	    char *msg;

	    if (errno == EPERM)
		msg = "root privilege required";
	    else
		msg = "internal error\n";

	    if (opt_debug)
		fprintf(stderr, "%s: %s\n", progname, msg);
	    else
		syslog(LOG_ERR, "%s", msg);

	    cleanup_and_exit(1);
	}
    }
}

/*
 * check_time
 *
 * Print the current time according to the host.
 */
void check_time()
{
    time_t ht = hosttime();

    printf("Host time: %s", ctime(&ht));
}

/*
 * filename_from_path
 *
 * Return a pointer to the filename part of a pathname.
 */
char *filename_from_path(char *pathname)
{
    char *ret = strrchr(pathname, '/');

    if (ret == NULL)
	ret = pathname;
    else
	ret++;

    return ret;
}

/*
 * usage
 *
 * Print help about the available options.
 */
void usage()
{
    fprintf(stderr, "vpctimed version %s\n\n", VERSION);
    fprintf(stderr, "Usage: %s [OPTION] ...\n\n", progname);

    fprintf(stderr,
	"  -c or --check             Print the time according to the host OS and exit.\n");
    fprintf(stderr,
	"  -f<n> or --fast=<n>       Adjust guest time if it is ahead by at least n\n");
    fprintf(stderr,
	"                            seconds.  The default is %d.\n", -DEFAULT_NEG_THRESHOLD);
    fprintf(stderr,
	"  -i<n> or --interval=<n>   Check and (potentially) adjust time every n\n");
    fprintf(stderr,
	"                            seconds.  The default is %d.\n", DEFAULT_INTERVAL);
    fprintf(stderr,
	"  -? or --help              Print this help information.\n");
    fprintf(stderr,
	"  -s<n> or --slow=<n>       Adjust guest time if it is behind by at least n\n");
    fprintf(stderr,
	"                            seconds.  The default is %d.\n", DEFAULT_POS_THRESHOLD);
    fprintf(stderr,
	"  -d or --debug             Print informational messages and don't detach\n");
    fprintf(stderr,
        "                            from terminal.\n");
}

int main(int argc, char *argv[])
{
    /* Get the name of the program. */
    progname = filename_from_path(argv[0]);

    /* Parse the command line arguments. */
    for (;;) {
	static struct option long_options[] = {
	    { "check", 0, NULL, 'c' },
	    { "debug", 0, NULL, 'd' },
	    { "fast", 1, NULL, 'f' },
	    { "help", 0, NULL, '?' },
	    { "interval", 1, NULL, 'i' },
	    { "slow", 1, NULL, 's' },
	    { 0, 0, 0, 0 }
	};

	int ch, longindex;
	char *tmpstr;

	ch = getopt_long(argc, argv, "?cdf:i:s:", long_options, &longindex);

	if (ch == -1)
	    break;

	switch (ch) {
	  case '?':
	    usage();
	    exit(0);
	  case 'c':
	    check_time();
	    exit(0);
	  case 'd':
	    opt_debug = TRUE;
	    break;
	  case 'f':
	    opt_neg_threshold = -strtod(optarg, &tmpstr);
	    if (tmpstr == optarg) {
		usage();
		exit(1);
	    }
	    break;
	  case 'i':
	    opt_interval = (int)strtol(optarg, &tmpstr, 10);
	    if (tmpstr == optarg) {
		usage();
		exit(1);
	    }
	    break;
	  case 's':
	    opt_pos_threshold = strtod(optarg, &tmpstr);
	    if (tmpstr == optarg) {
		usage();
		exit(1);
	    }
	    break;
	  default:
	    usage();
	    exit(1);
	}
    }

    if (optind < argc) {
	usage();
	exit(1);
    }

    /*
     * Check to see if the processor supports Connectix Virtual PC-specific
     * opcodes.  We set up a signal handler for illegal instructions, then
     * call hosttime() to use the opcodes.  If they are considered illegal by
     * the processor (in other words, if it is a real processor and not
     * Virtual PC), then the signal handler will print a message and exit.
     */
    signal(SIGILL, handle_sigill);
    hosttime();
    signal(SIGILL, SIG_DFL);

    /* We need root privilege to set the time - check if we have it. */
    if (geteuid() != 0) {
	fprintf(stderr, "%s: root privilege required\n", progname);
	exit(1);
    }

    if (opt_debug) {
	printf("Check clock every %d seconds.\n", opt_interval);
	printf("Tolerance: %.0f seconds slow - %.0f seconds fast.\n",
		opt_pos_threshold, -opt_neg_threshold);
    } else {
	FILE *f;

	/* Detach from the controlling terminal. */
	if (daemon(0,0) == -1) {
	    fprintf(stderr, "%s: unable to detach from terminal\n", argv[0]);
	    exit(1);
	}

	/* Write out a pid file (for startup/shutdown scripts). */
	if ((f = fopen(PID_FILE, "w")) != NULL) {
	    fprintf(f, "%d\n", getpid());
	    fclose(f);
	}
    }

    if (!opt_debug)
	openlog("vpctimed",
		(opt_debug ? LOG_PERROR : 0) | LOG_CONS, LOG_DAEMON);

    /* Catch termination signals. */
    signal(SIGINT, handle_signals);
    signal(SIGQUIT, handle_signals);
    signal(SIGTERM, handle_signals);

    /* Okay, we're all ready. */
    if (!opt_debug)
        syslog(LOG_INFO, "Started");

    /* Main loop. */
    for (;;) {
	synctime();
	sleep(opt_interval);
    }
}
